import datetime
import numpy as np
from optimus.tests.base import TestBase
from optimus.helpers.json import json_encoding
from optimus.helpers.functions import deep_sort, df_dicts_equal, results_equal


def Timestamp(t):
    return datetime.datetime.strptime(t, "%Y-%m-%d %H:%M:%S")


NaT = np.datetime64('NaT')
nan = float("nan")
inf = float("inf")


class TestMoreStatisticsPandas(TestBase):
    config = {'engine': 'pandas'}
    dict = {('id', 'int64'): [1, 2, 3, 4, 5, 6, 7, 8, 9, 10], ('name', 'object'): ['pants', 'shoes', 'shirt', 'pants', 'pants', 'shoes', 'pants', 'pants', 'shirt', 'pants'], ('code', 'object'): ['L15', 'SH', 'RG30', 'J10', 'JG15', 'B', 'JG20', 'L20', 'FT50', 'JG15'], ('price', 'float64'): [173.47, 69.99, 30.0, 34.99, 132.99, 57.99, 179.99, 95.0, 50.0, 169.99], ('discount', 'object'): ['0', '15%', '5%', '0', '0', '20%', '15%', '0', '0', '0']}
    maxDiff = None

    def test_cols_boxplot_all(self):
        df = self.df.copy()
        result = df.cols.boxplot(cols='*')
        expected = {'id': {'mean': {'id': 5.5, 'name': nan, 'code': nan, 'price': 99.441, 'discount': 0.0}, 'median': 5.5, 'q1': 3.25, 'q3': 7.75, 'whisker_low': -3.5, 'whisker_high': 14.5, 'fliers': [], 'label': 'id'}, 'name': nan, 'code': nan, 'price': {'mean': {'id': 5.5, 'name': nan, 'code': nan, 'price': 99.441, 'discount': 0.0}, 'median': 82.495, 'q1': 51.9975, 'q3': 160.74, 'whisker_low': -111.11625000000001, 'whisker_high': 323.85375, 'fliers': [], 'label': 'price'}, 'discount': {'mean': {'id': 5.5, 'name': nan, 'code': nan, 'price': 99.441, 'discount': 0.0}, 'median': 0.0, 'q1': 0.0, 'q3': 0.0, 'whisker_low': 0.0, 'whisker_high': 0.0, 'fliers': [], 'label': 'discount'}}
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_boxplot_multiple(self):
        df = self.df.copy()
        result = df.cols.boxplot(cols=['id', 'code', 'discount'])
        expected = {'id': {'mean': {'id': 5.5, 'code': nan, 'discount': 0.0}, 'median': 5.5, 'q1': 3.25, 'q3': 7.75, 'whisker_low': -3.5, 'whisker_high': 14.5, 'fliers': [], 'label': 'id'}, 'code': nan, 'discount': {'mean': {'id': 5.5, 'code': nan, 'discount': 0.0}, 'median': 0.0, 'q1': 0.0, 'q3': 0.0, 'whisker_low': 0.0, 'whisker_high': 0.0, 'fliers': [], 'label': 'discount'}}
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_boxplot_numeric(self):
        df = self.df.copy()
        result = df.cols.boxplot(cols='price')
        expected = {'price': {'mean': 99.441, 'median': 82.495, 'q1': 51.9975, 'q3': 160.74, 'whisker_low': -111.11625000000001, 'whisker_high': 323.85375, 'fliers': [], 'label': 'price'}}
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_correlation_all_kendall(self):
        df = self.df.copy()
        result = df.cols.correlation('*', 'kendall')
        expected = {'id': {'id': 1.0, 'price': 0.1111111111111111}, 'price': {'id': 0.1111111111111111, 'price': 1.0}}
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_correlation_all_pearson(self):
        df = self.df.copy()
        result = df.cols.correlation('*', 'pearson')
        expected = {'id': {'id': 1.0, 'price': 0.15785706335886504}, 'price': {'id': 0.15785706335886504, 'price': 1.0}}
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_correlation_all_spearman(self):
        df = self.df.copy()
        result = df.cols.correlation('*', 'spearman')
        expected = {'id': {'id': 1.0, 'price': 0.1393939393939394}, 'price': {'id': 0.1393939393939394, 'price': 1.0}}
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_correlation_multiple_kendall(self):
        df = self.df.copy()
        result = df.cols.correlation(['id', 'price'], 'kendall')
        expected = 0.1111111111111111
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_correlation_multiple_pearson(self):
        df = self.df.copy()
        result = df.cols.correlation(['id', 'price'], 'pearson')
        expected = 0.15785706335886504
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_correlation_multiple_spearman(self):
        df = self.df.copy()
        result = df.cols.correlation(['id', 'price'], 'spearman')
        expected = 0.1393939393939394
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_correlation_numeric_kendall(self):
        df = self.df.copy()
        result = df.cols.correlation('price', 'kendall')
        expected = {'price': {'price': 1.0}}
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_correlation_numeric_pearson(self):
        df = self.df.copy()
        result = df.cols.correlation('price', 'pearson')
        expected = {'price': {'price': 1.0}}
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_correlation_numeric_spearman(self):
        df = self.df.copy()
        result = df.cols.correlation('price', 'spearman')
        expected = {'price': {'price': 1.0}}
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_count_uniques_all(self):
        df = self.df.copy()
        result = df.cols.count_uniques(cols='*')
        expected = {'id': 10, 'name': 3, 'code': 9, 'price': 10, 'discount': 4}
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_count_uniques_multiple(self):
        df = self.df.copy()
        result = df.cols.count_uniques(cols=['id', 'code', 'discount'], estimate=False)
        expected = {'id': 10, 'code': 9, 'discount': 4}
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_count_uniques_numeric(self):
        df = self.df.copy()
        result = df.cols.count_uniques(cols='price', estimate=True)
        expected = 10
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_count_zeros_all(self):
        df = self.df.copy()
        result = df.cols.count_zeros(cols='*')
        expected = {'id': 0, 'name': 0, 'code': 0, 'price': 0, 'discount': 0}
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_count_zeros_multiple(self):
        df = self.df.copy()
        result = df.cols.count_zeros(cols=['id', 'code', 'discount'])
        expected = {'id': 0, 'code': 0, 'discount': 0}
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_count_zeros_numeric(self):
        df = self.df.copy()
        result = df.cols.count_zeros(cols='price')
        expected = 0
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_frequency_all(self):
        df = self.create_dataframe(data={('vf', 'float64'): [9.9, 9.9, 9.9, 9.9, 3.3000000000000003, 3.3000000000000003, 3.3000000000000003, 22.0, 22.0, 1.1, 1.1, 0.0, 0.0, 4.4], ('vs', 'object'): ['STR9', 'STR9', 'STR9', 'STR9', 'STR3', 'STR3', 'STR3', 'STR20', 'STR20', 'STR1', 'STR1', 'STR0', 'STR0', 'STR4'], ('values', 'int64'): [9, 9, 9, 9, 3, 3, 3, 20, 20, 1, 1, 0, 0, 4], ('o', 'object'): ['nine', [9], {'nine': 9}, 9, 3, 3, 3, None, None, 1, 1, 1, 1, 1]}, force_data_types=True)
        result = df.cols.frequency(cols='*', n=10, count_uniques=True)
        expected = {'frequency': {'vf': {'values': [{'value': 9.9, 'count': 4}, {'value': 3.3000000000000003, 'count': 3}, {'value': 0.0, 'count': 2}, {'value': 1.1, 'count': 2}, {'value': 22.0, 'count': 2}, {'value': 4.4, 'count': 1}], 'count_uniques': 6}, 'vs': {'values': [{'value': 'STR9', 'count': 4}, {'value': 'STR3', 'count': 3}, {'value': 'STR0', 'count': 2}, {'value': 'STR1', 'count': 2}, {'value': 'STR20', 'count': 2}, {'value': 'STR4', 'count': 1}], 'count_uniques': 6}, 'values': {'values': [{'value': 9, 'count': 4}, {'value': 3, 'count': 3}, {'value': 0, 'count': 2}, {'value': 1, 'count': 2}, {'value': 20, 'count': 2}, {'value': 4, 'count': 1}], 'count_uniques': 6}, 'o': {'values': [{'value': 1, 'count': 5}, {'value': 3, 'count': 3}, {'value': 9, 'count': 1}, {'value': [9], 'count': 1}, {'value': 'nine', 'count': 1}, {'value': {'nine': 9}, 'count': 1}], 'count_uniques': 6}}}
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_frequency_multiple(self):
        df = self.create_dataframe(data={('vf', 'float64'): [9.9, 9.9, 9.9, 9.9, 3.3000000000000003, 3.3000000000000003, 3.3000000000000003, 22.0, 22.0, 1.1, 1.1, 0.0, 0.0, 4.4], ('vs', 'object'): ['STR9', 'STR9', 'STR9', 'STR9', 'STR3', 'STR3', 'STR3', 'STR20', 'STR20', 'STR1', 'STR1', 'STR0', 'STR0', 'STR4'], ('values', 'int64'): [9, 9, 9, 9, 3, 3, 3, 20, 20, 1, 1, 0, 0, 4], ('o', 'object'): ['nine', [9], {'nine': 9}, 9, 3, 3, 3, None, None, 1, 1, 1, 1, 1]}, force_data_types=True)
        result = df.cols.frequency(cols=['vs', 'vf'], n=6, percentage=True)
        expected = {'frequency': {'vs': {'values': [{'value': 'STR9', 'count': 4, 'percentage': 28.57}, {'value': 'STR3', 'count': 3, 'percentage': 21.43}, {'value': 'STR0', 'count': 2, 'percentage': 14.29}, {'value': 'STR1', 'count': 2, 'percentage': 14.29}, {'value': 'STR20', 'count': 2, 'percentage': 14.29}, {'value': 'STR4', 'count': 1, 'percentage': 7.14}]}, 'vf': {'values': [{'value': 9.9, 'count': 4, 'percentage': 28.57}, {'value': 3.3000000000000003, 'count': 3, 'percentage': 21.43}, {'value': 0.0, 'count': 2, 'percentage': 14.29}, {'value': 1.1, 'count': 2, 'percentage': 14.29}, {'value': 22.0, 'count': 2, 'percentage': 14.29}, {'value': 4.4, 'count': 1, 'percentage': 7.14}]}}}
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_frequency_numeric(self):
        df = self.create_dataframe(data={('vf', 'float64'): [9.9, 9.9, 9.9, 9.9, 3.3000000000000003, 3.3000000000000003, 3.3000000000000003, 22.0, 22.0, 1.1, 1.1, 0.0, 0.0, 4.4], ('vs', 'object'): ['STR9', 'STR9', 'STR9', 'STR9', 'STR3', 'STR3', 'STR3', 'STR20', 'STR20', 'STR1', 'STR1', 'STR0', 'STR0', 'STR4'], ('values', 'int64'): [9, 9, 9, 9, 3, 3, 3, 20, 20, 1, 1, 0, 0, 4], ('o', 'object'): ['nine', [9], {'nine': 9}, 9, 3, 3, 3, None, None, 1, 1, 1, 1, 1]}, force_data_types=True)
        result = df.cols.frequency(cols='values', n=4, percentage=True, total_rows=3)
        expected = {'frequency': {'values': {'values': [{'value': 9, 'count': 4, 'percentage': 28.57}, {'value': 3, 'count': 3, 'percentage': 21.43}, {'value': 1, 'count': 2, 'percentage': 14.29}, {'value': 20, 'count': 2, 'percentage': 14.29}]}}}
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_frequency_string(self):
        df = self.create_dataframe(data={('vf', 'float64'): [9.9, 9.9, 9.9, 9.9, 3.3000000000000003, 3.3000000000000003, 3.3000000000000003, 22.0, 22.0, 1.1, 1.1, 0.0, 0.0, 4.4], ('vs', 'object'): ['STR9', 'STR9', 'STR9', 'STR9', 'STR3', 'STR3', 'STR3', 'STR20', 'STR20', 'STR1', 'STR1', 'STR0', 'STR0', 'STR4'], ('values', 'int64'): [9, 9, 9, 9, 3, 3, 3, 20, 20, 1, 1, 0, 0, 4], ('o', 'object'): ['nine', [9], {'nine': 9}, 9, 3, 3, 3, None, None, 1, 1, 1, 1, 1]}, force_data_types=True)
        result = df.cols.frequency(cols='vs', n=5, percentage=False)
        expected = {'frequency': {'vs': {'values': [{'value': 'STR9', 'count': 4}, {'value': 'STR3', 'count': 3}, {'value': 'STR0', 'count': 2}, {'value': 'STR1', 'count': 2}, {'value': 'STR20', 'count': 2}]}}}
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_heatmap_numeric_numeric(self):
        df = self.df.copy()
        result = df.cols.heatmap(col_x='discount', col_y='price', bins_x=5, bins_y=10)
        expected = {'x': {'name': 'discount', 'edges': [-0.5, 0.5]}, 'y': {'name': 'price', 'edges': [34.99, 173.47]}, 'values': [[0.0, 0.0, 1.0, 0.0, 0.0], [0.0, 0.0, 1.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0, 0.0], [0.0, 0.0, 1.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0, 0.0], [0.0, 0.0, 1.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0, 0.0], [0.0, 0.0, 2.0, 0.0, 0.0]]}
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_heatmap_numeric_string(self):
        df = self.df.copy()
        result = df.cols.heatmap(col_x='price', col_y='code', bins_x=3, bins_y=1)
        expected = {'x': {'name': 'price', 'edges': [0.0, 1.0]}, 'y': {'name': 'code', 'edges': [0.0, 1.0]}, 'values': [[0.0, 0.0, 0.0]]}
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_heatmap_string_numeric(self):
        df = self.df.copy()
        result = df.cols.heatmap(col_x='name', col_y='id', bins_x=7, bins_y=10)
        expected = {'x': {'name': 'name', 'edges': [0.0, 1.0]}, 'y': {'name': 'id', 'edges': [0.0, 1.0]}, 'values': [[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]]}
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_heatmap_string_string(self):
        df = self.df.copy()
        result = df.cols.heatmap(col_x='code', col_y='name', bins_x=4, bins_y=4)
        expected = {'x': {'name': 'code', 'edges': [0.0, 1.0]}, 'y': {'name': 'name', 'edges': [0.0, 1.0]}, 'values': [[0.0, 0.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0]]}
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_hist_all(self):
        df = self.df.copy()
        result = df.cols.hist(cols='*', buckets=2)
        expected = {'hist': {'id': [{'lower': 1.0, 'upper': 5.5, 'count': 5}, {'lower': 5.5, 'upper': 10.0, 'count': 5}], 'price': [{'lower': 30.0, 'upper': 104.995, 'count': 6}, {'lower': 104.995, 'upper': 179.99, 'count': 4}], 'discount': [{'lower': 0.0, 'upper': 0.0, 'count': 6}]}}
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_hist_multiple(self):
        df = self.df.copy()
        result = df.cols.hist(cols=['id', 'code', 'discount'], buckets=4)
        expected = {'hist': {'id': [{'lower': 1.0, 'upper': 3.25, 'count': 3}, {'lower': 3.25, 'upper': 5.5, 'count': 2}, {'lower': 5.5, 'upper': 7.75, 'count': 2}, {'lower': 7.75, 'upper': 10.0, 'count': 3}], 'discount': [{'lower': 0.0, 'upper': 0.0, 'count': 6}]}}
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_hist_numeric(self):
        df = self.df.copy()
        result = df.cols.hist(cols='price', buckets=10)
        expected = {'hist': {'price': [{'lower': 30.0, 'upper': 44.999, 'count': 2}, {'lower': 44.999, 'upper': 59.998000000000005, 'count': 2}, {'lower': 59.998000000000005, 'upper': 74.997, 'count': 1}, {'lower': 74.997, 'upper': 89.99600000000001, 'count': 0}, {'lower': 89.99600000000001, 'upper': 104.995, 'count': 1}, {'lower': 104.995, 'upper': 119.994, 'count': 0}, {'lower': 119.994, 'upper': 134.993, 'count': 1}, {'lower': 134.993, 'upper': 149.99200000000002, 'count': 0}, {'lower': 149.99200000000002, 'upper': 164.991, 'count': 0}, {'lower': 164.991, 'upper': 179.99, 'count': 3}]}}
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_infer_type_all(self):
        df = self.df.copy()
        result = df.cols.infer_type(cols='*')
        expected = {'id': {'data_type': 'int', 'categorical': True}, 'name': {'data_type': 'str', 'categorical': True}, 'code': {'data_type': 'str', 'categorical': True}, 'price': {'data_type': 'float', 'categorical': True}, 'discount': {'data_type': 'int', 'categorical': True}}
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_infer_type_multiple(self):
        df = self.df.copy()
        result = df.cols.infer_type(cols=['id', 'code', 'discount'])
        expected = {'id': {'data_type': 'int', 'categorical': True}, 'code': {'data_type': 'str', 'categorical': True}, 'discount': {'data_type': 'int', 'categorical': True}}
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_infer_type_numeric(self):
        df = self.df.copy()
        result = df.cols.infer_type(cols='price')
        expected = {'data_type': 'float', 'categorical': True}
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_profile_all(self):
        df = self.create_dataframe(data={('vf', 'float64'): [9.9, 9.9, 9.9, 9.9, 3.3000000000000003, 3.3000000000000003, 3.3000000000000003, 22.0, 22.0, 1.1, 1.1, 0.0, 0.0, 4.4], ('vs', 'object'): ['STR9', 'STR9', 'STR9', 'STR9', 'STR3', 'STR3', 'STR3', 'STR20', 'STR20', 'STR1', 'STR1', 'STR0', 'STR0', 'STR4'], ('values', 'int64'): [9, 9, 9, 9, 3, 3, 3, 20, 20, 1, 1, 0, 0, 4], ('o', 'object'): ['nine', [9], {'nine': 9}, 9, 3, 3, 3, None, None, 1, 1, 1, 1, 1]}, force_data_types=True)
        result = df.cols.profile(cols='*')
        expected = [{'stats': {'match': 14, 'missing': 0, 'mismatch': 0, 'inferred_data_type': {'data_type': 'float', 'categorical': True}, 'frequency': [{'value': 9.9, 'count': 4}, {'value': 3.3000000000000003, 'count': 3}, {'value': 0.0, 'count': 2}, {'value': 1.1, 'count': 2}, {'value': 22.0, 'count': 2}, {'value': 4.4, 'count': 1}], 'count_uniques': 6}, 'data_type': 'float64'}, {'stats': {'match': 14, 'missing': 0, 'mismatch': 0, 'inferred_data_type': {'data_type': 'str', 'categorical': True}, 'frequency': [{'value': 'STR9', 'count': 4}, {'value': 'STR3', 'count': 3}, {'value': 'STR0', 'count': 2}, {'value': 'STR1', 'count': 2}, {'value': 'STR20', 'count': 2}, {'value': 'STR4', 'count': 1}], 'count_uniques': 6}, 'data_type': 'object'}, {'stats': {'match': 14, 'missing': 0, 'mismatch': 0, 'inferred_data_type': {'data_type': 'int', 'categorical': True}, 'frequency': [{'value': 9, 'count': 4}, {'value': 3, 'count': 3}, {'value': 0, 'count': 2}, {'value': 1, 'count': 2}, {'value': 20, 'count': 2}, {'value': 4, 'count': 1}], 'count_uniques': 6}, 'data_type': 'int64'}, {'stats': {'match': 9, 'missing': 2, 'mismatch': 3, 'inferred_data_type': {'data_type': 'int', 'categorical': True}, 'frequency': [{'value': 1, 'count': 5}, {'value': 3, 'count': 3}, {'value': 9, 'count': 1}, {'value': [9], 'count': 1}, {'value': 'nine', 'count': 1}, {'value': {'nine': 9}, 'count': 1}], 'count_uniques': 6}, 'data_type': 'object'}]
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_profile_multiple(self):
        df = self.create_dataframe(data={('vf', 'float64'): [9.9, 9.9, 9.9, 9.9, 3.3000000000000003, 3.3000000000000003, 3.3000000000000003, 22.0, 22.0, 1.1, 1.1, 0.0, 0.0, 4.4], ('vs', 'object'): ['STR9', 'STR9', 'STR9', 'STR9', 'STR3', 'STR3', 'STR3', 'STR20', 'STR20', 'STR1', 'STR1', 'STR0', 'STR0', 'STR4'], ('values', 'int64'): [9, 9, 9, 9, 3, 3, 3, 20, 20, 1, 1, 0, 0, 4], ('o', 'object'): ['nine', [9], {'nine': 9}, 9, 3, 3, 3, None, None, 1, 1, 1, 1, 1]}, force_data_types=True)
        result = df.cols.profile(cols=['vs', 'vf'], bins=8, flush=False)
        expected = [{'stats': {'match': 14, 'missing': 0, 'mismatch': 0, 'inferred_data_type': {'data_type': 'str', 'categorical': True}, 'frequency': [{'value': 'STR9', 'count': 4}, {'value': 'STR3', 'count': 3}, {'value': 'STR0', 'count': 2}, {'value': 'STR1', 'count': 2}, {'value': 'STR20', 'count': 2}, {'value': 'STR4', 'count': 1}], 'count_uniques': 6}, 'data_type': 'object'}, {'stats': {'match': 14, 'missing': 0, 'mismatch': 0, 'inferred_data_type': {'data_type': 'float', 'categorical': True}, 'frequency': [{'value': 9.9, 'count': 4}, {'value': 3.3000000000000003, 'count': 3}, {'value': 0.0, 'count': 2}, {'value': 1.1, 'count': 2}, {'value': 22.0, 'count': 2}, {'value': 4.4, 'count': 1}], 'count_uniques': 6}, 'data_type': 'float64'}]
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_profile_numeric(self):
        df = self.create_dataframe(data={('vf', 'float64'): [9.9, 9.9, 9.9, 9.9, 3.3000000000000003, 3.3000000000000003, 3.3000000000000003, 22.0, 22.0, 1.1, 1.1, 0.0, 0.0, 4.4], ('vs', 'object'): ['STR9', 'STR9', 'STR9', 'STR9', 'STR3', 'STR3', 'STR3', 'STR20', 'STR20', 'STR1', 'STR1', 'STR0', 'STR0', 'STR4'], ('values', 'int64'): [9, 9, 9, 9, 3, 3, 3, 20, 20, 1, 1, 0, 0, 4], ('o', 'object'): ['nine', [9], {'nine': 9}, 9, 3, 3, 3, None, None, 1, 1, 1, 1, 1]}, force_data_types=True)
        result = df.cols.profile(cols='values', bins=10, flush=True)
        expected = {'stats': {'match': 14, 'missing': 0, 'mismatch': 0, 'inferred_data_type': {'data_type': 'int', 'categorical': True}, 'frequency': [{'value': 9, 'count': 4}, {'value': 3, 'count': 3}, {'value': 0, 'count': 2}, {'value': 1, 'count': 2}, {'value': 20, 'count': 2}, {'value': 4, 'count': 1}], 'count_uniques': 6}, 'data_type': 'int64'}
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_quality_all(self):
        df = self.df.copy()
        result = df.cols.quality(cols='*')
        expected = {'id': {'match': 10, 'missing': 0, 'mismatch': 0, 'inferred_data_type': 'int'}, 'name': {'match': 10, 'missing': 0, 'mismatch': 0, 'inferred_data_type': 'str'}, 'code': {'match': 10, 'missing': 0, 'mismatch': 0, 'inferred_data_type': 'str'}, 'price': {'match': 10, 'missing': 0, 'mismatch': 0, 'inferred_data_type': 'float'}, 'discount': {'match': 6, 'missing': 0, 'mismatch': 4, 'inferred_data_type': 'int'}}
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_quality_multiple(self):
        df = self.df.copy()
        result = df.cols.quality(cols=['id', 'code', 'discount'], flush=False)
        expected = {'id': {'match': 10, 'missing': 0, 'mismatch': 0, 'inferred_data_type': 'int'}, 'code': {'match': 10, 'missing': 0, 'mismatch': 0, 'inferred_data_type': 'str'}, 'discount': {'match': 6, 'missing': 0, 'mismatch': 4, 'inferred_data_type': 'int'}}
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_quality_numeric(self):
        df = self.df.copy()
        result = df.cols.quality(cols='price', flush=True)
        expected = {'price': {'match': 10, 'missing': 0, 'mismatch': 0, 'inferred_data_type': 'float'}}
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_unique_values_all(self):
        df = self.df.copy()
        result = df.cols.unique_values(cols='*')
        expected = {'id': ['1', '2', '3', '4', '5', '6', '7', '8', '9', '10'], 'name': ['pants', 'shoes', 'shirt'], 'code': ['L15', 'SH', 'RG30', 'J10', 'JG15', 'B', 'JG20', 'L20', 'FT50'], 'price': ['173.47', '69.99', '30.0', '34.99', '132.99', '57.99', '179.99', '95.0', '50.0', '169.99'], 'discount': ['0', '15%', '5%', '20%']}
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_unique_values_multiple(self):
        df = self.df.copy()
        result = df.cols.unique_values(cols=['id', 'code', 'discount'], estimate=False)
        expected = {'id': ['1', '2', '3', '4', '5', '6', '7', '8', '9', '10'], 'code': ['L15', 'SH', 'RG30', 'J10', 'JG15', 'B', 'JG20', 'L20', 'FT50'], 'discount': ['0', '15%', '5%', '20%']}
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_unique_values_numeric(self):
        df = self.df.copy()
        result = df.cols.unique_values(cols='price', estimate=True)
        expected = ['173.47', '69.99', '30.0', '34.99', '132.99', '57.99', '179.99', '95.0', '50.0', '169.99']
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))


class TestMoreStatisticsDask(TestMoreStatisticsPandas):
    config = {'engine': 'dask', 'n_partitions': 1}


class TestMoreStatisticsPartitionDask(TestMoreStatisticsPandas):
    config = {'engine': 'dask', 'n_partitions': 2}


try:
    import cudf # pyright: reportMissingImports=false
except:
    pass
else:
    class TestMoreStatisticsCUDF(TestMoreStatisticsPandas):
        config = {'engine': 'cudf'}


try:
    import dask_cudf # pyright: reportMissingImports=false
except:
    pass
else:
    class TestMoreStatisticsDC(TestMoreStatisticsPandas):
        config = {'engine': 'dask_cudf', 'n_partitions': 1}


try:
    import dask_cudf # pyright: reportMissingImports=false
except:
    pass
else:
    class TestMoreStatisticsPartitionDC(TestMoreStatisticsPandas):
        config = {'engine': 'dask_cudf', 'n_partitions': 2}


try:
    import pyspark # pyright: reportMissingImports=false
except:
    pass
else:
    class TestMoreStatisticsSpark(TestMoreStatisticsPandas):
        config = {'engine': 'spark'}


try:
    import vaex # pyright: reportMissingImports=false
except:
    pass
else:
    class TestMoreStatisticsVaex(TestMoreStatisticsPandas):
        config = {'engine': 'vaex'}



from optimus.helpers.columns import parse_columns
from optimus.plots.functions import plot_heatmap, plot_boxplot, plot_frequency, plot_hist, \
    plot_correlation, plot_qqplot


class Plot:
    def __init__(self, df):
        self.df = df

    def hist(self, columns=None, buckets=10, output_format="plot", output_path=None):
        """
        Plot histogram
        :param columns: Columns to be printed
        :param buckets: Number of buckets
        :param output_format:
        :param output_path: path where the image is going to be saved
        :return:
        """
        df = self.df
        columns = parse_columns(df, columns)

        data = df.cols.hist(columns, buckets)["hist"]
        for col_name in data.keys():
            plot_hist({col_name: data[col_name]}, output=output_format, path=output_path)

    def heatmap(self, col_x, col_y, bins_x=10, bins_y=10, output_format="plot", output_path=None):
        """
        Plot heat map
        :param col_x: columns to be printed
        :param bins_x: number of buckets
        :param output_format:
        :param output_path: path where the image is going to be saved
        :return:
        """
        df = self.df
        data = df.cols.heatmap(col_x, col_y, bins_x, bins_y)

        plot_heatmap(data, output=output_format, path=output_path)

    def box(self, columns=None, output_format="plot", output_path=None):
        """
        Plot boxplot
        :param columns: Columns to be printed
        :param output_format:
        :param output_path: path where the image is going to be saved
        :return:
        """
        stats = self.df.cols.boxplot(columns)
        plot_boxplot(stats, output=output_format, path=output_path)

    def frequency(self, columns=None, buckets=10, output_format="plot", output_path=None):
        """
        Plot frequency chart
        :param columns: Columns to be printed
        :param buckets: Number of buckets
        :param output_format:
        :param output_path: path where the image is going to be saved
        :return:
        """
        df = self.df
        columns = parse_columns(df, columns)
        data = df.cols.frequency(columns, buckets, tidy=True)

        for k, v in data.items():
            plot_frequency({k: v}, output=output_format, path=output_path)

    def correlation(self, col_name, method="pearson", output_format="plot", output_path=None):
        """
        Compute the correlation matrix for the input data set of Vectors using the specified method. Method
        mapped from pyspark.ml.stat.Correlation.
        :param col_name: The name of the column for which the correlation coefficient needs to be computed.
        :param method: String specifying the method to use for computing correlation. Supported: pearson (default),
        spearman.
        :param output_format: Output image format
        :param output_path: Output path
        :return: Heatmap plot of the corr matrix using seaborn.
        """
        df = self.df
        cols_data = df.cols.correlation(col_name, method)
        plot_correlation(cols_data, output=output_format, path=output_path)

    def qqplot(self, columns, n=100, output_format="plot", output_path=None):
        """
        QQ plot
        :param columns:
        :param n: Sample size
        :param output_format: Output format
        :param output_path: Path to the output file
        :return:
        """
        df = self.df

        columns = parse_columns(
            df, cols_args=columns, filter_by_column_types=df.constants.NUMERIC_TYPES)

        if columns is not None:
            sample_data = df.sample(n=n, random=True)
            for col_name in columns:
                plot_qqplot(col_name, sample_data, output=output_format, path=output_path)



from abc import ABC, abstractmethod

import pprint

from optimus.helpers.columns import parse_columns
from optimus.helpers.core import val_to_list, one_list_to_val
from optimus.helpers.filters import dict_filter
from optimus.helpers.json import dump_json


class AbstractOutlierBounds(ABC):
    """
     This is a template class to expand the outliers methods
     Also you need to add the new outlier detection method to outliers.py
     """

    def __init__(self, df, col_name: str, lower_bound: int, upper_bound: int):
        """

        :param df: Spark Dataframe
        :param col_name: column name
        """
        self.df = df
        self.col_name = one_list_to_val(parse_columns(df, col_name))
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound

    def __repr__(self):
        return pprint.pformat(self.info())

    @abstractmethod
    def whiskers(self):
        """
        Get the whiskers and IQR
        :return:
        """
        pass

    def select(self):
        """
        Select outliers rows using the selected column
        :return:
        """
        df = self.df
        col_name = self.col_name
        upper_bound, lower_bound = dict_filter(self.whiskers(), ["upper_bound", "lower_bound"])
        return df.rows.select((df[col_name] > upper_bound) | (df[col_name] < lower_bound))

    # TODO: Pass a defined division param instead or run 3 separated jobs
    def hist(self, col_name: str):
        df = self.df
        # lower bound
        lower_bound_hist = df.rows.select(df[col_name] < self.lower_bound).cols.hist(col_name, 20)

        # upper bound
        upper_bound_hist = df.rows.select(df[col_name] > self.upper_bound).cols.hist(col_name, 20)

        # Non outliers
        non_outlier_hist = df.rows.select(
            (df[col_name] >= self.lower_bound) & (df[col_name] <= self.upper_bound)).cols.hist(col_name, 20)

        result = {}
        if lower_bound_hist is not None:
            result.update(lower_bound_hist)
        if upper_bound_hist is not None:
            result.update(upper_bound_hist)
        if non_outlier_hist is not None:
            result.update(non_outlier_hist)

        return dump_json(result)

    def select_lower_bound(self):
        col_name = self.col_name
        df = self.df
        return df.rows.select(df[col_name] <= self.lower_bound)

    def select_upper_bound(self):
        col_name = self.col_name
        df = self.df
        return df.rows.select(df[col_name] >= self.upper_bound)

    def drop(self):
        """
        Drop outliers rows using the selected column
        :return:
        """

        df = self.df
        col_name = self.col_name
        # upper_bound, lower_bound = dict_filter(self.whiskers(), ["upper_bound", "lower_bound"])
        return df.rows.drop((df[col_name] > self.upper_bound) | (df[col_name] < self.lower_bound))

    def count_lower_bound(self, bound: int):
        """
        Count outlier in the lower bound
        :return:
        """
        col_name = self.col_name
        df = self.df
        return df.rows.select(df[col_name] < bound).rows.count()

    def count_upper_bound(self, bound: int):
        """
        Count outliers in the upper bound
        :return:
        """
        col_name = self.col_name
        df = self.df
        return df.rows.select(df[col_name] > bound).rows.count()

    def count(self):
        """
        Count the outliers rows using the selected column
        :return:
        """
        col_name = self.col_name
        df = self.df
        return df.rows.select((df[col_name] > self.upper_bound) | (df[col_name] < self.lower_bound)).rows.count()

    def non_outliers_count(self):
        """
        Count non outliers rows using the selected column
        :return:
        """
        df = self.df
        col_name = self.col_name
        return df.rows.select((df[col_name] <= self.upper_bound) & (df[col_name] >= self.lower_bound)).rows.count()

    @abstractmethod
    def info(self, output: str = "dict"):
        """
        Get whiskers, iqrs and outliers and non outliers count
        :return:
        """
        pass



import copy
from itertools import combinations_with_replacement
from optimus.helpers.types import *

import jellyfish as jellyfish
import numpy as np

from optimus.engines.base.ml.constants import CLUSTER_COL
from optimus.helpers.columns import parse_columns, name_col
from optimus.helpers.output import output_json
from optimus.helpers.raiseit import RaiseIt
from optimus.infer import is_list, is_str


class Clusters:
    type = None
    clusters = {}

    def __init__(self, clusters):
        self.clusters = clusters

    def __repr__(self):
        return str(output_json(self.to_dict(verbose=True)))

    def set_suggestions(self, suggestions, column=0):

        for i, suggestion in enumerate(suggestions):
            self.set_suggestion(i, suggestion, column)

    def set_suggestion(self, suggestion_or_id, new_value, column=0):

        if isinstance(column, (int,)):
            column = list(self.clusters.keys())[column]

        if is_str(suggestion_or_id):
            for i in range(len(self.clusters[column])):
                if (self.clusters[column][i]["suggestion"] == suggestion_or_id) or (
                        self.clusters[column][i]["cluster"] == suggestion_or_id):
                    suggestion_or_id = i
                    break

            if is_str(suggestion_or_id):
                raise ValueError(f"'{suggestion_or_id}' is not a cluster of '{column}'")

        self.clusters[column][suggestion_or_id]["suggestion"] = new_value

    def to_dict(self, columns="*", limit_clusters=None, limit_suggestions=None, verbose=False):
        result = {}

        columns = self._parse_columns(columns)

        for column in columns:
            result[column] = {}
            clusters = self.clusters[column]
            if not is_list(clusters):
                clusters = list(clusters.values())
                
            for cluster in clusters[0:limit_clusters]:
                if verbose:
                    cluster_name = cluster["cluster"]
                    _cluster = copy.deepcopy(cluster)
                    del _cluster["cluster"]
                    result[column][cluster_name] = _cluster
                else:
                    cluster_name = cluster["suggestion"]
                    result[column][cluster_name] = result[column].get(cluster_name, [])
                    result[column][cluster_name] += cluster["suggestions"][0:limit_suggestions]

        return result

    def _parse_columns(self, columns):

        if columns == "*":
            columns = list(self.clusters.keys())
        else:
            for column in columns:
                if not self.clusters.get(column, False) and isinstance(column, (int,)):
                    column = list(self.clusters.keys())[column]
                    if not self.clusters.get(column, False):
                        raise

        return columns

    def display(self, columns="*", limit_clusters=None, limit_suggestions=None, verbose=True):
        return self.to_dict(columns, limit_clusters, limit_suggestions, verbose)


def string_clustering(df, cols="*", algorithm=None, *args, **kwargs) -> 'ClustersType':
    """
    Cluster a dataframe column based on the Fingerprint algorithm
    :return:
    """

    funcs = ["fingerprint", "ngram_fingerprint", "metaphone", "nysiis", "match_rating_codex",
             "double_metaphone", "soundex", "levenshtein"]

    if algorithm in funcs:
        func = getattr(df.cols, algorithm)
    else:
        RaiseIt.value_error(algorithm, funcs)

    cols = parse_columns(df, cols)
    result = {}

    for input_col in cols:
        if algorithm == "levenshtein":
            a = np.empty((0, 3))
            _df = df[input_col].rows.drop_duplicated(how="all")
            _df = df.cols.fingerprint(input_col)

            for b, z in zip(combinations_with_replacement(_df.data[input_col], 2),
                            combinations_with_replacement(_df.data[input_col], 2)):
                a = np.append(a, np.array([[b[0], b[1], jellyfish.levenshtein_distance(*z)]], dtype="object"), axis=0)

            total_rows = df.rows.count()
            topn = 2
            result = {}
            suggestions = []

            for s in _df.data[input_col]:
                b = a[((a[:, 0] == s) | (a[:, 1] == s))]

                c = b[b[:, 2].argsort()][:topn][:, [0, 1]]
                c = [s] + c[np.where(c != s)].tolist()
                suggestions.append({
                    "cluster": s,
                    "suggestion": c[0], 
                    "suggestions": c, 
                    "suggestions_size": len(c), 
                    "total_count": total_rows
                })
            result[input_col] = suggestions
        else:

            cluster_col = name_col(input_col, CLUSTER_COL)
            # TODO: Think we can remove duplicated values to improve time calculation performace
            df = func(input_col, output_cols=cluster_col, *args, **kwargs)

            # sets the original values as indices: {"VALUE": "value", "Value": "value"}

            _dfd = df.cols.select([input_col, cluster_col]).cols.to_string().data.set_index(input_col)
            _df = df.new(_dfd)
            values = _df[cluster_col].to_pandas().to_dict()[cluster_col]

            suggestions_items = {}
            for k, v in values.items():
                suggestions_items[v] = suggestions_items.get(v, []) + [k]

            counts_list = df.cols.frequency(cluster_col, len(values))['frequency'][cluster_col]['values']
            suggestions = []

            for d in counts_list:
                value = d['value']
                suggestion = {"suggestion": value, "cluster": value, "total_count": d['count']}
                if suggestions_items[value]:
                    suggestion["suggestions"] = suggestions_items[value]
                    suggestion["suggestion"] = suggestion["suggestions"][0]
                    suggestion["suggestions_size"] = len(suggestions_items[value])

                suggestions += [suggestion]

            result[input_col] = suggestions
    return Clusters(result)



from enum import Enum, unique


@unique
class DriverProperties(Enum):
    """Enum holding all driver properties."""

    REDIS = {
        "name": "redis",
        "port": 6379,
        "table_name": "",
        "java_class":""
    }

    CASSANDRA = {
        "name": "cassandra",
        "port": 9042,
        "table_name": "",
        "java_class": ""
    }
    MYSQL = {
        "name": "mysql",
        "port": 3306,
        "table_name": "table_name",
        "java_class": ""
    }
    ORACLE = {
        "name": "oracle",
        "port": 1521,
        "table_name": "table_name",
        "java_class": ""
    }
    POSTGRESQL = {
        "name": "postgresql",
        "port": 5432,
        "table_name": "table_name",
        "java_class": ""
    }
    SQLSERVER = {
        "name": "mssql+pyodbc",
        "port": 1433,
        "table_name": "TABLE_NAME",
        "java_class": ""
    }
    SQLITE = {
        "name": "sqlite",
        "port": 0,
        "table_name": "name",
        "java_class": ""
    }
    PRESTO = {
        "name": "presto",
        "port": 8080,
        "table_name": "",
        "java_class": ""
    }
    REDSHIFT = {
        "name": "redshift",
        "port": 5439,
        "table_name": "table_name",
        "java_class": ""
    }
    BIGQUERY = {
        "name": "bigquery",
        "port": 5439,
        "table_name": "table_name",
        "java_class": ""
    }
    IMPALA = {
        "name": "impala",
        "port": 10000,
        "table_name": "table_name",
        "java_class": ""
    }


    @classmethod
    def list(cls):
        return list(map(lambda c: c.value, cls))



from optimus.engines.base.dask.dataframe import DaskBaseDataFrame
from optimus.engines.base.cudf.dataframe import CUDFBaseDataFrame
from optimus.engines.cudf.dataframe import CUDFDataFrame
from optimus.engines.dask_cudf.io.save import Save
from optimus.engines.pandas.dataframe import PandasDataFrame
from optimus.helpers.columns import parse_columns
from optimus.helpers.converter import cudf_to_dask_cudf


class DaskCUDFDataFrame(CUDFBaseDataFrame, DaskBaseDataFrame):

    def _base_to_dfd(self, pdf, n_partitions):
        return cudf_to_dask_cudf(pdf, n_partitions)

    @property
    def rows(self):
        from optimus.engines.dask_cudf.rows import Rows
        return Rows(self)

    @property
    def cols(self):
        from optimus.engines.dask_cudf.columns import Cols
        return Cols(self)

    @property
    def functions(self):
        from optimus.engines.dask_cudf.functions import DaskCUDFFunctions
        return DaskCUDFFunctions(self)

    @property
    def save(self):
        return Save(self)

    @property
    def mask(self):
        from optimus.engines.dask_cudf.mask import DaskCUDFMask
        return DaskCUDFMask(self)

    @property
    def constants(self):
        from optimus.engines.base.dask.constants import Constants
        return Constants()

    def _buffer_window(self, input_cols, lower_bound, upper_bound):
        return PandasDataFrame(self.get_buffer().data[input_cols][lower_bound: upper_bound].to_pandas().reset_index(drop=True), op=self.op, label_encoder=self.le)

    @staticmethod
    def melt(id_vars, value_vars, var_name="variable", value_name="value", data_type="str"):
        pass

    @staticmethod
    def query(sql_expression):
        pass

    @staticmethod
    def debug():
        pass

    def to_pandas(self):
        return self.data.compute().to_pandas()

    def to_optimus_pandas(self):
        return PandasDataFrame(self.root.to_pandas(), op=self.op)

    def to_optimus_cudf(self):
        return CUDFDataFrame(self.root.to_pandas(), op=self.op)



from optimus.tests.base import TestBase


class TestLoadPandas(TestBase):

    def test_json(self):
        df = self.load_dataframe("examples/data/foo.json", type="json", multiline=True)
        self.assertEqual(df.rows.count(), 19)
        self.assertEqual(df.cols.names(), ["id", "firstName", "lastName", "billingId", "product", "price", "birth", "dummyCol"])
        self.assertEqual(self.config.get("n_partitions", 1), df.partitions())

    def test_json_less_rows(self):
        df = self.load_dataframe("examples/data/foo.json", type="json", n_rows=13)
        self.assertEqual(df.rows.count(), 13)
        self.assertEqual(df.cols.names(), ["id", "firstName", "lastName", "billingId", "product", "price", "birth", "dummyCol"])
        self.assertEqual(self.config.get("n_partitions", 1), df.partitions())

    def test_json_more_rows(self):
        df = self.load_dataframe("examples/data/foo.json", type="json", n_rows=50)
        self.assertLess(df.rows.count(), 50)
        self.assertEqual(df.cols.names(), ["id", "firstName", "lastName", "billingId", "product", "price", "birth", "dummyCol"])
        self.assertEqual(self.config.get("n_partitions", 1), df.partitions())

    def test_xml(self):
        df = self.load_dataframe("examples/data/foo.xml", type="xml")
        self.assertEqual(df.rows.count(), 19)
        self.assertEqual(df.cols.names(), ["id", "firstName", "lastName", "billingId", "product", "price", "birth", "dummyCol"])
        self.assertEqual(self.config.get("n_partitions", 1), df.partitions())

    def test_xml_less_rows(self):
        df = self.load_dataframe("examples/data/foo.xml", type="xml", n_rows=13)
        self.assertEqual(df.rows.count(), 13)
        self.assertEqual(df.cols.names(), ["id", "firstName", "lastName", "billingId", "product", "price", "birth", "dummyCol"])
        self.assertEqual(self.config.get("n_partitions", 1), df.partitions())

    def test_xml_more_rows(self):
        df = self.load_dataframe("examples/data/foo.xml", type="xml", n_rows=50)
        self.assertLess(df.rows.count(), 50)
        self.assertEqual(df.cols.names(), ["id", "firstName", "lastName", "billingId", "product", "price", "birth", "dummyCol"])
        self.assertEqual(self.config.get("n_partitions", 1), df.partitions())

    def test_parquet(self):
        df = self.load_dataframe("examples/data/foo.parquet", type="parquet")
        self.assertEqual(df.rows.count(), 19)
        self.assertEqual(df.cols.names(), ["id", "firstName", "lastName", "billingId", "product", "price", "birth", "dummyCol"])
        self.assertEqual(self.config.get("n_partitions", 1), df.partitions())

    def test_parquet_less_rows(self):
        df = self.load_dataframe("examples/data/foo.parquet", type="parquet", n_rows=13)
        self.assertEqual(df.rows.count(), 13)
        self.assertEqual(df.cols.names(), ["id", "firstName", "lastName", "billingId", "product", "price", "birth", "dummyCol"])
        self.assertEqual(self.config.get("n_partitions", 1), df.partitions())

    def test_parquet_more_rows(self):
        df = self.load_dataframe("examples/data/foo.parquet", type="parquet", n_rows=50)
        self.assertLess(df.rows.count(), 50)
        self.assertEqual(df.cols.names(), ["id", "firstName", "lastName", "billingId", "product", "price", "birth", "dummyCol"])
        self.assertEqual(self.config.get("n_partitions", 1), df.partitions())

    def test_avro(self):
        df = self.load_dataframe("examples/data/foo.avro", type="avro")
        self.assertEqual(df.rows.count(), 19)
        self.assertEqual(df.cols.names(), ["id", "firstName", "lastName", "billingId", "product", "price", "birth", "dummyCol"])
        self.assertEqual(self.config.get("n_partitions", 1), df.partitions())

    def test_avro_less_rows(self):
        df = self.load_dataframe("examples/data/foo.avro", type="avro", n_rows=13)
        self.assertEqual(df.rows.count(), 13)
        self.assertEqual(df.cols.names(), ["id", "firstName", "lastName", "billingId", "product", "price", "birth", "dummyCol"])
        self.assertEqual(self.config.get("n_partitions", 1), df.partitions())

    def test_avro_more_rows(self):
        df = self.load_dataframe("examples/data/foo.avro", type="avro", n_rows=50)
        self.assertLess(df.rows.count(), 50)
        self.assertEqual(df.cols.names(), ["id", "firstName", "lastName", "billingId", "product", "price", "birth", "dummyCol"])
        self.assertEqual(self.config.get("n_partitions", 1), df.partitions())

    def test_tsv(self):
        df = self.load_dataframe("examples/data/foo.tsv", type="tsv")
        self.assertEqual(df.rows.count(), 5)
        self.assertEqual(df.cols.names(), ["Sepal length", "Sepal width", "Petal length", "Petal width", "Species"])
        self.assertEqual(self.config.get("n_partitions", 1), df.partitions())

    def test_tsv_less_rows(self):
        df = self.load_dataframe("examples/data/foo.tsv", type="tsv", n_rows=3)
        self.assertEqual(df.rows.count(), 3)
        self.assertEqual(df.cols.names(), ["Sepal length", "Sepal width", "Petal length", "Petal width", "Species"])        
        self.assertEqual(self.config.get("n_partitions", 1), df.partitions())

    def test_tsv_more_rows(self):
        df = self.load_dataframe("examples/data/foo.tsv", type="tsv", n_rows=50)
        self.assertLess(df.rows.count(), 50)
        self.assertEqual(df.cols.names(), ["Sepal length", "Sepal width", "Petal length", "Petal width", "Species"])        
        self.assertEqual(self.config.get("n_partitions", 1), df.partitions())

    def test_xls(self):
        df = self.load_dataframe("examples/data/titanic3.xls", type="excel")
        self.assertEqual(df.rows.count(), 1309)
        self.assertEqual(df.cols.names(), ["pclass", "survived", "name", "sex", "age", "sibsp", "parch", "ticket", "fare", "cabin", "embarked", "boat", "body", "home.dest"])
        self.assertEqual(self.config.get("n_partitions", 1), df.partitions())

    def test_xls_less_rows(self):
        df = self.load_dataframe("examples/data/titanic3.xls", type="excel", n_rows=13)
        self.assertEqual(df.rows.count(), 13)
        self.assertEqual(df.cols.names(), ["pclass", "survived", "name", "sex", "age", "sibsp", "parch", "ticket", "fare", "cabin", "embarked", "boat", "body", "home.dest"])
        self.assertEqual(self.config.get("n_partitions", 1), df.partitions())

    def test_xls_more_rows(self):
        df = self.load_dataframe("examples/data/titanic3.xls", type="excel", n_rows=50)
        self.assertLess(df.rows.count(), 5000)
        self.assertEqual(df.cols.names(), ["pclass", "survived", "name", "sex", "age", "sibsp", "parch", "ticket", "fare", "cabin", "embarked", "boat", "body", "home.dest"])
        self.assertEqual(self.config.get("n_partitions", 1), df.partitions())

class TestLoadDask(TestLoadPandas):
    config = {'engine': 'dask', 'n_partitions': 1}


class TestLoadPartitionDask(TestLoadPandas):
    config = {'engine': 'dask', 'n_partitions': 2}


try:
    import cudf # pyright: reportMissingImports=false
except:
    pass
else:
    class TestLoadCUDF(TestLoadPandas):
        config = {'engine': 'cudf'}


try:
    import dask_cudf # pyright: reportMissingImports=false
except:
    pass
else:
    class TestLoadDC(TestLoadPandas):
        config = {'engine': 'dask_cudf', 'n_partitions': 1}


try:
    import dask_cudf # pyright: reportMissingImports=false
except:
    pass
else:
    class TestLoadPartitionDC(TestLoadPandas):
        config = {'engine': 'dask_cudf', 'n_partitions': 2}


try:
    import pyspark
except:
    pass
else:
    class TestLoadSpark(TestLoadPandas):
        config = {'engine': 'spark'}

try:
    import vaex
except:
    pass
else:
    class TestLoadVaex(TestLoadPandas):
        config = {'engine': 'vaex'}


from optimus.server.functions import run, code, engine, session, create_session, delete_session, features

class Engine:
    
    def __init__(self, config=None, session=None, **kwargs):

        if session is not None:
            self.id = session.id
        else:
            self.id = "default"
        
        if isinstance(config, str):
            engineConfig = {"engine": config}
            config = None
        else:
            engineConfig = {}
        
        engineConfig.update(config or kwargs)
        e = engine(self.id, engineConfig)
        self.name = e["updated"]

    def request(self, body=None, **kwargs):
        body = body or kwargs
        if "source" not in body:
            body.update({"source": self.name})
        return run(self.id, body)


from optimus.engines.base.basedataframe import BaseDataFrame
import numpy as np

class Functions(type):
    """
    Handle the method name and execute the function corresponding to the dataframe class
    """
    def __getattr__(self, func) -> callable:
        """

        :param func: Method name
        :return:
        """

        def _func(df_or_value, *args):
            # Get the internal series in the optimus dataframe
            if isinstance(df_or_value, BaseDataFrame):
                _df = df_or_value.data[df_or_value.cols.names(0)[0]]
                # Convert the frame to a dataframe and to a optimus Dataframe
                series = getattr(df_or_value.functions, func)(_df, *args)
                series.name = df_or_value.cols.names(0)[0]
                r = df_or_value.__class__(series.to_frame(), df_or_value.op)
            else:
                r = getattr(np, func)(df_or_value, *args)

            if isinstance(r, np.generic):
                return np.asscalar(r)
            return r

        return _func


class F(metaclass=Functions):
    """
    Dummy class to handle method names
    For example: In F.sqrt, sqrt is captured and mapped a matched to the function dataframe
    """
    pass



import timeit
from functools import wraps

from optimus.helpers.logger import logger


def time_it(method):
    def timed(*args, **kw):
        start_time = timeit.default_timer()
        f = method(*args, **kw)
        _time = round(timeit.default_timer() - start_time, 2)
        logger.print("{name}() executed in {time} sec".format(name=method.__name__, time=_time))
        return f

    return timed



from abc import ABC
from optimus.helpers.functions import match_date
from optimus.engines.base.meta import Meta
from optimus.helpers.types import *

from optimus.helpers.columns import parse_columns
from optimus.helpers.constants import ProfilerDataTypes
from optimus.helpers.core import val_to_list, one_list_to_val
from optimus.helpers.raiseit import RaiseIt
from optimus.infer import is_dict, is_str, regex_http_code, regex_social_security_number, regex_phone_number, \
    regex_credit_card_number, regex_zip_code, regex_gender, regex_ip, regex_email, \
    is_datetime, is_list, is_bool, is_object, regex_url


class Mask(ABC):
    def __init__(self, root: 'DataFrameType'):
        self.root = root
        self.F = root.functions

    def _to_frame(self, series):
        if callable(getattr(series, "to_frame", False)):
            return series.to_frame()
        return series

    def numeric(self, cols="*"):
        return self.root[cols].cols.apply(cols, self.F.is_numeric)

    def int(self, cols="*"):
        return self.root[cols].cols.apply(cols, self.F.is_integer)

    def float(self, cols="*"):
        return self.root[cols].cols.apply(cols, self.F.is_float)

    def str(self, cols="*"):
        return self.root[cols].cols.apply(cols, self.F.is_string)

    def greater_than(self, cols="*", value=None) -> 'MaskDataFrameType':
        df = self.root
        return df[cols] > value

    def greater_than_equal(self, cols="*", value=None) -> 'MaskDataFrameType':
        df = self.root
        return df[cols] >= value

    def less_than(self, cols="*", value=None) -> 'MaskDataFrameType':
        df = self.root
        return df[cols] < value

    def less_than_equal(self, cols="*", value=None) -> 'MaskDataFrameType':
        df = self.root
        return df[cols] <= value

    def between(self, cols="*", lower_bound=None, upper_bound=None, equal=True, bounds=None) -> 'MaskDataFrameType':
        
        dfc = self.root[cols]
        
        if bounds is None:
            if lower_bound is None and upper_bound is None:
                RaiseIt.type_error((lower_bound, upper_bound), (int, float))
            else:
                bounds = [(lower_bound, upper_bound)]

        mask = None

        for lower_b, upper_b in bounds:
            
            if upper_b is not None and lower_b is not None:

                upper_b = float(upper_b) if is_str(upper_b) else upper_b
                lower_b = float(lower_b) if is_str(lower_b) else lower_b

                if upper_b < lower_b:
                    upper_b, lower_b = lower_b, upper_b

                if equal:
                    _mask = (dfc >= lower_b) & (dfc <= upper_b)
                else:
                    _mask = (dfc > lower_b) & (dfc < upper_b)
            else:
                if equal:
                    _mask = (dfc >= lower_b) if lower_b is not None else (dfc <= upper_b)
                else:
                    _mask = (dfc > lower_b) if lower_b is not None else (dfc < upper_b)

            mask = _mask if mask is None else (mask | _mask)

        return mask


    def equal(self, cols="*", value=None) -> 'MaskDataFrameType':
        df = self.root
        return df[cols] == value

    def not_equal(self, cols="*", value=None) -> 'MaskDataFrameType':
        df = self.root
        return df[cols] != value

    def missing(self, cols="*") -> 'MaskDataFrameType':
        """
        Return missing values
        :param cols:
        :return:
        """
        return self.null(cols) | self.empty(cols)

    def mismatch(self, cols="*", data_type=None) -> 'MaskDataFrameType':
        """
        Return missing values
        :param cols:
        :param data_type:
        :return:
        """

        df = self.root

        if is_dict(cols) and data_type is None:
            data_type = [col["data_type"] for key, col in cols.items()]
            cols = [col for col in cols]

        cols = one_list_to_val(parse_columns(df, cols))
        data_type = one_list_to_val(data_type)

        mask_match = df[cols].mask.match_data_type(cols, data_type)
        mask_null = df[cols].mask.null(cols)
        return ~(mask_match | mask_null)

    def match(self, cols="*", arg=None, regex=None, data_type=None) -> 'MaskDataFrameType':

        if arg is not None:
            if arg in ProfilerDataTypes.list():
                data_type = arg
            else:
                regex = arg

        if data_type is None:
            return self.match_regex(cols=cols, regex=regex)
        else:
            return self.match_data_type(cols=cols, data_type=data_type)

    def match_regex(self, cols="*", regex="") -> 'MaskDataFrameType':
        return self.root[cols].cols.apply(cols, self.root.cols.F.match, args=(regex,))

    def match_data_type(self, cols="*", data_type=None) -> 'MaskDataFrameType':
        """
        Return values that match with a datatype
        :param cols:
        :param data_type:
        :return:
        """
        df = self.root
        cols = one_list_to_val(parse_columns(df, cols))

        if data_type is None:
            inferred_data_type = df.profile.data_type(cols)
            if inferred_data_type is None:
                RaiseIt.value_error(data_type, ProfilerDataTypes.list(),
                                    "data_type not found in cache. Need to passed as data_type param")
            else:
                data_type = inferred_data_type

        if is_list(data_type):
            mask_match = None
            for _col, _data_type in zip(cols, data_type):
                _data_type = df.constants.INTERNAL_TO_OPTIMUS.get(_data_type, _data_type)
                if mask_match is None:
                    mask_match = getattr(df[_col].mask, _data_type)(_col)
                else:
                    mask_match[_col] = getattr(df[_col].mask, _data_type)(_col)
        else:
            data_type = df.constants.INTERNAL_TO_OPTIMUS.get(data_type, data_type)
            mask_match = getattr(df[cols].mask, data_type)(cols)

        return mask_match

    def value_in(self, cols="*", values=None) -> 'MaskDataFrameType':
        
        df = self.root
        cols = parse_columns(df, cols)

        values = val_to_list(values)

        mask = df.data[cols].isin(values)
        return df.new(self._to_frame(mask))

    def pattern(self, cols="*", pattern=None) -> 'MaskDataFrameType':
        
        df = self.root
        cols = parse_columns(df, cols)

        return df[cols].cols.pattern() == pattern

    def starts_with(self, cols="*", value=None) -> 'MaskDataFrameType':
                
        df = self.root
        cols = val_to_list(parse_columns(df, cols))

        mask = None

        for col in cols:
            series = df.functions.to_string_accessor(df.data[col]).startswith(value, na=False)
            if mask is None:
                mask = self._to_frame(series)
            else:
                mask[col] = series

        return df.new(self._to_frame(mask))

    def ends_with(self, cols="*", value=None) -> 'MaskDataFrameType':
                
        df = self.root
        cols = val_to_list(parse_columns(df, cols))

        mask = None

        for col in cols:
            series = df.functions.to_string_accessor(df.data[col]).endswith(value, na=False)
            if mask is None:
                mask = self._to_frame(series)
            else:
                mask[col] = series

        return df.new(self._to_frame(mask))

    def contains(self, cols="*", value=None, case=True, flags=0, na=False, regex=False) -> 'MaskDataFrameType':
        """

        :param cols:
        :param value:
        :param case:
        :param flags:
        :param na:
        :param regex:
        :return:
        """
                
        df = self.root
        cols = val_to_list(parse_columns(df, cols))

        mask = None

        if is_list(value):
            value = "|".join(value)
            regex = True

        for col in cols:
            series = df.functions.to_string_accessor(df.data[col]).contains(value, case=case, flags=flags, na=na, regex=regex)
            if mask is None:
                mask = self._to_frame(series)
            else:
                mask[col] = series

        return df.new(self._to_frame(mask))

    def expression(self, where=None, cols="*") -> 'MaskDataFrameType':
        """
        Find rows and appends resulting mask to the dataset
        :param where: Mask, expression or name of the column to be taken as mask
        :param cols:
        :return: Optimus Dataframe
        """

        df = self.root.cols.select(cols)

        if is_str(where):
            if where in df.cols.names():
                where = df[where]
            else:
                where = eval(where)

        if isinstance(where, (df.__class__,)):
            return where.cols.to_boolean()

        RaiseIt.type_error(where, ["Dataframe", "Expression", "Column name"])

    def find(self, cols="*", value=None) -> 'MaskDataFrameType':
        
        df = self.root
        cols = val_to_list(parse_columns(df, cols))

        if is_str(value):

            mask = None
            for col in cols:
                series = df.functions.to_string_accessor(df.data[col]).match(value, na=False)
                if mask is None:
                    mask = self._to_frame(series)
                else:
                    mask[col] = series

            return self.root.new(self._to_frame(mask))

        else:
            return df[cols] == value



    def null(self, cols="*", how="any") -> 'MaskDataFrameType':
        """
        Find the rows that have null values
        :param how:
        :param cols:
        :return:
        """
        df = self.root
        dfd = self.root.data
        cols = val_to_list(parse_columns(df, cols))
        subset_df = dfd[cols]

        if how == "all":
            col_name = cols[0] if len(cols) == 1 else "__null__"
            mask = subset_df.isnull().all(axis=1).rename(col_name)
        else:
            mask = subset_df.isnull()

        return self.root.new(self._to_frame(mask))

    
    def none(self, cols="*") -> 'MaskDataFrameType':
        """
        Find the rows that have None values

        :param cols:
        :return:
        """
        return ~self.numeric(cols) & self.null(cols)

    def nan(self, cols="*") -> 'MaskDataFrameType':
        """
        Find the rows that have np.nan values

        :param cols:
        :return:
        """
        return self.numeric(cols) & self.null(cols)

    def duplicated(self, cols="*", keep="first") -> 'MaskDataFrameType':
        """
        Find the rows that have duplicated values

        :param keep:
        :param cols:
        :return:
        """
        df = self.root
        dfd = self.root.data
        cols = val_to_list(parse_columns(df, cols))

        if cols is not None:
            subset = cols
            subset_df = dfd[subset]
        else:
            subset_df = dfd

        col_name = cols[0] if len(cols) == 1 else "__duplicated__"

        object_cols = [col for col in subset_df.columns if subset_df[col].dtype.name in df.constants.OBJECT_TYPES]

        subset_df[object_cols] = subset_df[object_cols].astype(str)

        mask = self.F.duplicated(subset_df, keep, cols).rename(col_name)

        return self.root.new(self._to_frame(mask))

    def unique(self, cols, keep="first") -> 'MaskDataFrameType':
        """
        Find the rows that are not duplicated values

        :param keep:
        :param cols:
        :return:
        """
        return ~self.duplicated(cols, keep)

    def empty(self, cols="*") -> 'MaskDataFrameType':
        """
        Find the rows that do not have any info

        :param cols:
        :return:
        """
        return self.root[cols] == ""

    def email(self, cols="*") -> 'MaskDataFrameType':
        return self.match_regex(cols, regex_email)

    def ip(self, cols="*") -> 'MaskDataFrameType':
        return self.match_regex(cols, regex_ip)

    def url(self, cols="*", schemes="default", allow_schemeless=True) -> 'MaskDataFrameType':
        """
        Find the rows with urls
        :param cols:
        :param schemes: 
        :param allow_schemeless:
        """

        if schemes == "default":
            schemes = ['http', 'https', 'ftp', 'ftps']

        if schemes == "any":
            schemes_regex = r"[a-zA-Z0-9]*:\/\/"
        elif schemes is not None:
            schemes = val_to_list(schemes)
            schemes_regex = r"(" + "|".join(schemes) + r"):\/\/"
        else:
            schemes_regex = ""
            allow_schemeless = False

        if allow_schemeless:
            schemes_regex = r"(" + schemes_regex + r")?"
        
        return self.match_regex(cols, r"^" + schemes_regex + regex_url)

    def gender(self, cols="*") -> 'MaskDataFrameType':
        return self.match_regex(cols, regex_gender)

    def boolean(self, cols="*") -> 'MaskDataFrameType':
        return self.root[cols].cols.apply(cols, is_bool)

    def zip_code(self, cols="*") -> 'MaskDataFrameType':
        return self.match_regex(cols, regex_zip_code)

    def credit_card_number(self, cols="*") -> 'MaskDataFrameType':
        return self.match_regex(cols, regex_credit_card_number)

    def datetime(self, cols="*") -> 'MaskDataFrameType':
        df = self.root
        dtypes = list(set(df.constants.DATETIME_TYPES) - set(df.constants.ANY_TYPES))
        datetime_cols = parse_columns(df, cols, filter_by_column_types=dtypes)
        cols = parse_columns(df, cols)
        non_datetime_cols = list(set(cols) - set(datetime_cols)) if datetime_cols else cols
        df = df[cols]
        if datetime_cols and len(datetime_cols):
            df = df.cols.apply(datetime_cols, is_datetime)
        if non_datetime_cols and len(non_datetime_cols):

            date_formats = df.cols.date_format(non_datetime_cols, tidy=False)["date_format"]

            for col_name, date_format in date_formats.items():
                if date_format is True:
                    df = df.cols.assign({col_name: True})
                elif date_format:
                    regex = match_date(date_format)
                    if not regex:
                        df = df.cols.assign({col_name: False})
                    else:
                        df[col_name] = df.mask.match_regex(col_name, regex)
                else:
                    df = df.cols.assign({col_name: False})

        return df

    def object(self, cols="*") -> 'MaskDataFrameType':
        return self.root[cols].cols.apply(cols, is_object)
        # return self.match_regex(cols, is_object)

    def array(self, cols="*") -> 'MaskDataFrameType':
        return self.root[cols].cols.apply(cols, is_list)
        # return self.match_regex(cols, is_list_value)

    def phone_number(self, cols="*") -> 'MaskDataFrameType':
        return self.match_regex(cols, regex_phone_number)

    def social_security_number(self, cols="*") -> 'MaskDataFrameType':
        return self.match_regex(cols, regex_social_security_number)

    def http_code(self, cols="*") -> 'MaskDataFrameType':
        return self.match_regex(cols, regex_http_code)

    #

    def all(self, cols="*") -> 'MaskDataFrameType':
        
        df = self.root

        mask = None

        for col in df.cols.names(cols):
            _mask = df[col]
            if mask is None:
                mask = _mask
            else:
                mask = mask & _mask

        return mask

    def any(self, cols="*") -> 'MaskDataFrameType':
        
        df = self.root

        mask = None

        for col in df.cols.names(cols):
            _mask = df[col]
            if mask is None:
                mask = _mask
            else:
                mask = mask | _mask

        return mask
